#ifndef SET_H
#define SET_H

#include "Node.h"
#include <utility>

template <class T>
class Set {
public:
	Set();
	Set(const Set<T>& org);
	~Set();
	Set<T>& operator=(const Set<T>& rhs);
	std::pair<bool, Node<T>* > insert(const T& val);
	Node<T>* find(const T& val, const Node<T>* ptr=0) const;
	void remove(const T& val);
	Node<T>* begin() const;
	Node<T>* end() const;
private:
	unsigned int sz;
	Node<T>* root;
	Node<T> endNode;
	Node<T>* duplicate(const Node<T>* input, const Node<T>* par=0);
	void destroy(const Node<T>* nd);
	pair<bool, Node<T>*> add_node(const T &val, Node<T> *nptr, const Node<T> *prnt);
};

template <class T>
Set<T>::Set() : sz(0)
{
	root = &endNode;
}

template <class T>
Node<T>* Set<T>::begin() const
{
	return root;
}

template <class T>
Node<T>* Set<T>::end() const
{
	return &endNode;
}

template <class T>
Set<T>::Set(const Set<T>& org)
{
	root = duplicate(org.root);
}

template <class T>
Node<T>* Set<T>::duplicate(const Node<T>* input, const Node<T>* par)
{
	Node<T>* newNode = new Node<T>(input);
	newNode->parent = par;
	if(newNode->left!=&endNode && newNode->left!=0) {
		newNode->left = duplicate(input->left, input);
	}
	if(newNode->right!=&endNode && newNode->right!=0) {
		newNode->right = duplicate(input->right, input);
	}
	return newNode;
}

template <class T>
Set<T>::~Set()
{
	destroy(root);
}

template <class T>
Set<T>& Set<T>::operator=(const Set<T>& rhs)
{
	if(this!=&rhs) {
		destroy(root);
		sz = rhs.sz;
		endNode = rhs.endNode;
		root = duplicate(rhs.root);
	}
	return *this;
}

template <class T>
Node<T>* Set<T>::find(const T& val, Node<T>* ptr) const
{
	Node<T>* ret = root;
	if(ptr==0) {
		ptr = root;
	}
	if(ptr->value==val) {
		ret = ptr;
	} else if(val<ptr->value) {
		if(ptr->left==&endNode || ptr->left==0) {
			ret = &endNode;
		} else {
			ret = find(val, ptr->left);
		}
	} else if(ptr->right==&endNode || ptr->right==0) {
		ret = &endNode;
	} else {
		ret = find(val, ptr->right);
	}
	return ret;
}

template <class T>
std::pair<bool, Node<T>* > Set<T>::insert(const T& val)
{
		return add_node(val, root, NULL);
}

template <class T>
pair<bool, Node<T>*> Set<T>::add_node(const T &val, Node<T> *nptr, const Node<T> *prnt)
{
	pair<bool, Node<T>*> retVal;
	retVal.second = &endNode;
	retVal.first = false;
	
	if(nptr == NULL || nptr == &endNode)
	{
		nptr = new Node<T>(val, &endNode, &endNode, prnt);
		retVal.second = nptr;
		retVal.first = true;
		++sz;
	}
	if(val < nptr->value) {
		retVal = add_node(val, nptr->left, nptr); 
	}
	if(val > nptr->value){
		retVal = add_node(val, nptr->right, nptr); 
	}
	return retVal;
}

template <class T>
void Set<T>::remove(const T& val)
{
	// Incomplete
	Node<T>* soonGone = find(val);
	if(soonGone!=&endNode) {
		if(soonGone->parent->right == soonGone) {
			if(soonGone->left!=0 && soonGone->left!=&endNode) {
				Node<T>* parent = soonGone->parent;
				Node<T>* rightMostChild = rightMost(soonGone->left);
				parent->right = soonGone->left;
				soonGone->left->parent = parent;
				rightMostChild->right = soonGone->right;
				soonGone->right->parent = rightMostChild;
				delete soonGone;
				--sz;
			}
		} else {
			if(soonGone->right!=0 && soonGone->right!=&endNode) {
				Node<T>* parent = soonGone->parent;
				Node<T>* leftMostChild = leftMost(soonGone->right);
				parent->left = soonGone->right;
				soonGone->right->parent = parent;
				leftMostChild->left = soonGone->left;
				soonGone->left->parent = leftMostChild;
				delete soonGone;
				--sz;
			}
		}
	}
}

template <class T>
void Set<T>::destroy(const Node<T>* nd)
{
	if(nd->left!=&endNode && nd->left!=0) {
		destroy(nd->left);
	}
	if(nd->right!=&endNode && nd->right!=0) {
		destroy(nd->right);
	}
	delete nd;
}

#endif